/* IMWheel Configuration Helper Definitions
 * Copylefted under the Linux GNU Public License
 * Author : Jonathan Atkins <jcatki@jonatkins.org>
 * PLEASE: contact me if you have any improvements, I will gladly code good ones
 */
#ifndef CFG_H
#define CFG_H
#include <X11/X.h>

int cfg(int); /* wheel_fd or -1 */
#endif
